@@ -24,18 +24,22 @@
 
 #include "pgtclCmds.h"
 #include "pgtclId.h"
+#ifdef HAVE_SQLITE3
+#  include "pgtclSqlite.h"
+#endif
 
 static int
-PgEndCopy(Pg_ConnectionId * connid, int *errorCodePtr)
+PgEndCopy(Pg_ConnectionId * connid, int *errorCodePtr, int writing)
 {
 	connid->res_copyStatus = RES_COPY_NONE;
-	if (PQendcopy(connid->conn))
+	if (writing && PQputCopyEnd(connid->conn, NULL) != 1)
 	{
 		PQclear(connid->results[connid->res_copy]);
 		connid->results[connid->res_copy] =
 			PQmakeEmptyPGresult(connid->conn, PGRES_BAD_RESPONSE);
 		connid->res_copy = -1;
 		*errorCodePtr = EIO;
+		PgCheckConnectionState(connid);
 		return -1;
 	}
 	else
@@ -57,6 +61,7 @@ PgInputProc(DRIVER_INPUT_PROTO)
 	Pg_ConnectionId *connid;
 	PGconn	   *conn;
 	int			avail;
+	char	*bufPtr = NULL;
 
 	connid = (Pg_ConnectionId *) cData;
 	conn = connid->conn;
@@ -65,6 +70,7 @@ PgInputProc(DRIVER_INPUT_PROTO)
 	 PQresultStatus(connid->results[connid->res_copy]) != PGRES_COPY_OUT)
 	{
 		*errorCodePtr = EBUSY;
+		PgCheckConnectionState(connid);
 		return -1;
 	}
 
@@ -75,17 +81,34 @@ PgInputProc(DRIVER_INPUT_PROTO)
 	if (!PQconsumeInput(conn))
 	{
 		*errorCodePtr = EIO;
+		PgCheckConnectionState(connid);
 		return -1;
 	}
 
 	/* Move data from libpq's buffer to Tcl's. */
 
-	avail = PQgetlineAsync(conn, buf, bufSize);
+	switch(avail = PQgetCopyData(conn, &bufPtr, bufSize)) {
+	  case -2: {
+		*errorCodePtr = EIO;
+		PgCheckConnectionState(connid);
+		return -1;
+	  }
+	  case -1: {
+		// PgEndCopy calls PgCheckConnectionState if needed.
+		return PgEndCopy(connid, errorCodePtr, 0);
+	  }
+	}
 
-	if (avail < 0)
-	{
-		/* Endmarker detected, change state and return 0 */
-		return PgEndCopy(connid, errorCodePtr);
+	/* Should never happen */
+	if(avail < 0) {
+		*errorCodePtr = EIO;
+		PgCheckConnectionState(connid);
+		return -1;
+	}
+
+	if(bufPtr) {
+		memcpy (buf, bufPtr, avail);
+		PQfreemem(bufPtr);
 	}
 
 	return avail;
@@ -99,6 +122,8 @@ PgOutputProc(DRIVER_OUTPUT_PROTO)
 {
 	Pg_ConnectionId *connid;
 	PGconn	   *conn;
+	int         endcopy = 0;
+	int         writeLen;
 
 	connid = (Pg_ConnectionId *) cData;
 	conn = connid->conn;
@@ -106,23 +131,33 @@ PgOutputProc(DRIVER_OUTPUT_PROTO)
 	if (connid->res_copy < 0 ||
 	  PQresultStatus(connid->results[connid->res_copy]) != PGRES_COPY_IN)
 	{
+		PgCheckConnectionState(connid);
 		*errorCodePtr = EBUSY;
 		return -1;
 	}
 
-	if (PQputnbytes(conn, buf, bufSize))
-	{
-		*errorCodePtr = EIO;
-		return -1;
-	}
+	writeLen = bufSize;
 
 	/*
 	 * This assumes Tcl script will write the terminator line in a single
 	 * operation; maybe not such a good assumption?
 	 */
-	if (bufSize >= 3 && strncmp(&buf[bufSize - 3], "\\.\n", 3) == 0)
+	if (writeLen >= 3 && strncmp(&buf[writeLen - 3], "\\.\n", 3) == 0)
 	{
-		if (PgEndCopy(connid, errorCodePtr) == -1)
+		writeLen -= 3; // Don't write the terminator using the new API
+		endcopy = 1;
+	}
+
+	if (PQputCopyData(conn, buf, writeLen) < 0)
+	{
+		*errorCodePtr = EIO;
+		PgCheckConnectionState(connid);
+		return -1;
+	}
+
+	if (endcopy) {
+		// PgEndCopy calls PgCheckConnectionState
+		if (PgEndCopy(connid, errorCodePtr, 1) == -1)
 			return -1;
 	}
 	return bufSize;
@@ -289,7 +324,11 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
         "sendquery_prepared",  "null_value_string", "version", 
         "protocol", "param", "backendpid", "socket", 
 	"conndefaults",  "set_single_row_mode", "is_busy", "blocking",
-	"cancel_request", (char *)NULL
+	"cancel_request", "copy_complete",
+#ifdef HAVE_SQLITE3
+	"sqlite",
+#endif
+	(char *)NULL
     };
 
     enum options
@@ -302,7 +341,10 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 	SENDQUERY_PREPARED, NULL_VALUE_STRING, VERSION, 
 	PROTOCOL, PARAM, BACKENDPID, SOCKET,
 	CONNDEFAULTS, SET_SINGLE_ROW_MODE, ISBUSY, BLOCKING,
-	CANCELREQUEST
+	CANCELREQUEST, COPY_COMPLETE,
+#ifdef HAVE_SQLITE3
+	SQLITE3
+#endif
     };
 
     if (objc == 1 || objc > 25)
@@ -432,7 +474,7 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
                      Tcl_GetString(objvx[objvxi]));
             }
             */
-               
+
             idx += num;
             objvx[idx] = Tcl_NewStringObj(connid->id, -1);
             returnCode = Pg_execute(cData, interp, objc, objvx);
@@ -574,8 +616,7 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
         }
 	case CONNDEFAULTS:
 	{
-            returnCode = Pg_conndefaults(cData, interp, 1, objvx);
-	    break;
+            return Pg_conndefaults(cData, interp, 1, objvx);
 	}
 	
 	case SET_SINGLE_ROW_MODE:
@@ -605,6 +646,22 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
             returnCode = Pg_cancelrequest(cData, interp, objc, objvx);
 	    break;
 	}
+
+	case COPY_COMPLETE:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_copy_complete(cData, interp, objc, objvx);
+	    break;
+	}
+
+#ifdef HAVE_SQLITE3
+	case SQLITE3:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_sqlite(cData, interp, objc, objvx);
+	    break;
+	}
+#endif
     }
 	Tcl_DecrRefCount(objvx[idx]);
 	return returnCode;
@@ -649,8 +706,6 @@ PgResultCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[
     return Pg_result(cData, interp, objc + 1, objvx);
 }
 
-
-
 /*
  * Get back the connection from the Id
  */
@@ -687,24 +742,25 @@ PgGetConnectionId(Tcl_Interp *interp, const char *id, Pg_ConnectionId ** connid_
 int
 PgDelConnectionId(DRIVER_DEL_PROTO)
 {
-	Tcl_HashEntry *entry;
-	Tcl_HashSearch hsearch;
+	Tcl_HashEntry   *entry;
+	Tcl_HashSearch   hsearch;
 	Pg_ConnectionId *connid;
-	Pg_TclNotifies *notifies;
-	int			i;
-        Pg_resultid    *resultid;
+	Pg_TclNotifies  *notifies;
+	int              i;
+        Pg_resultid     *resultid;
+	int              allow_unregister;
 
 
 	connid = (Pg_ConnectionId *) cData;
 
 	for (i = 0; i < connid->res_max; i++)
 	{
-	    if (connid->results[i])
+		if (connid->results[i])
 		{
 			PQclear(connid->results[i]);
 
 
-            resultid = connid->resultids[i];
+			resultid = connid->resultids[i];
 
 			if (resultid != NULL) {
 				Tcl_DecrRefCount(resultid->str);
@@ -742,7 +798,9 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 	 * pending notify and connection-loss events.
 	 */
 	PgStopNotifyEventSource(connid, 1);
- 
+
+	/* Check if the connection has been broken in the background */
+	allow_unregister = PQsocket(connid->conn) >= 0;
 
 	/* Close the libpq connection too */
 	PQfinish(connid->conn);
@@ -764,10 +822,10 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 	 * Note we are not leaking a socket, since libpq closed that already.
 	 */
 
-	if (connid->notifier_channel != NULL && interp != NULL)
-        {
+	if (allow_unregister && connid->notifier_channel != NULL && interp != NULL)
+	{
 		Tcl_UnregisterChannel(NULL, connid->notifier_channel);
-         }
+	}
 
        /*
         * Clear any async result callback, if present.
@@ -801,7 +859,7 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 /*
  *----------------------------------------------------------------------
  *
- * PgResultId --
+ * PgSetResultId --
  *
  *    Find a slot for a new result id.  If the table is full, expand 
  *    it by a factor of 2.  However, do not expand past the hard max, 
@@ -916,6 +974,7 @@ getresid(Tcl_Interp *interp, const char *id, Pg_ConnectionId ** connid_p)
 
 	if (!(mark = strrchr(id, '.')))
 	{
+		Tcl_SetResult(interp, "Poorly formated result handle", TCL_STATIC);
 		return -1;
 	}
 	*mark = '\0';
@@ -937,7 +996,6 @@ getresid(Tcl_Interp *interp, const char *id, Pg_ConnectionId ** connid_p)
 
 	if (resid < 0 || resid >= connid->res_max || connid->results[resid] == NULL)
 	{
-		
 		Tcl_SetResult(interp, "Invalid result handle", TCL_STATIC);
 		return -1;
 	}
@@ -1078,7 +1136,6 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 	NotifyEvent *event = (NotifyEvent *) evPtr;
 	Pg_TclNotifies *notifies;
 	char	   *callback;
-	char	   *svcallback;
 
 	/* We classify SQL notifies as Tcl file events. */
 	if (!(flags & TCL_FILE_EVENTS))
@@ -1141,18 +1198,30 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 		if (callback == NULL)
 			continue;			/* nothing to do for this interpreter */
 
-		/*
-		 * We have to copy the callback string in case the user executes a
-		 * new pg_listen or pg_on_connection_loss during the callback.
+		/* Create a Tcl List Object containing the callback with the channel name 
+		 * (relname) and the PID of the notifying backend. This will also copy the callback
+		 * string in case the user executes a new pg_listen or pg_on_connection_loss
+		 * during the callback
 		 */
-		svcallback = (char *)ckalloc((unsigned)(strlen(callback) + 1));
-		strcpy(svcallback, callback);
+		Tcl_Obj *callbackList = Tcl_NewListObj(0, NULL);
+		Tcl_ListObjAppendElement(NULL, callbackList, Tcl_NewStringObj(callback, -1));
+		if (event->notify) {
+			Tcl_ListObjAppendElement(NULL, callbackList, Tcl_NewStringObj(event->notify->relname, -1));
+			Tcl_ListObjAppendElement(NULL, callbackList, Tcl_NewIntObj(event->notify->be_pid));
+			/* In case there is a payload, add it to the callback as a
+			 * further element to the list.
+			 */
+			if (event->notify->extra[0]) {
+				Tcl_ListObjAppendElement (NULL, callbackList,  Tcl_NewStringObj(event->notify->extra, -1));
+			}
+		}
 
 		/*
 		 * Execute the callback.
 		 */
+		Tcl_IncrRefCount(callbackList);
 		Tcl_Preserve((ClientData)interp);
-		if (Tcl_GlobalEval(interp, svcallback) != TCL_OK)
+		if (Tcl_EvalObjEx(interp, callbackList, TCL_EVAL_GLOBAL) != TCL_OK)
 		{
 			if (event->notify)
 				Tcl_AddErrorInfo(interp, "\n    (\"pg_listen\" script)");
@@ -1161,7 +1230,7 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 			Tcl_BackgroundError(interp);
 		}
 		Tcl_Release((ClientData)interp);
-		ckfree(svcallback);
+		Tcl_DecrRefCount(callbackList);
 
 		/*
 		 * Check for the possibility that the callback closed the
@@ -1500,3 +1569,70 @@ PgDelResultHandle(ClientData cData)
 
     return;
 }
+
+/**********************************
+ * pg_copy_complete
+ *
+ * complete a COPY IN operation
+ *
+ * Usage: pg_copy_complete connection
+ *
+ **********************************/
+
+int
+Pg_copy_complete(ClientData cData, Tcl_Interp *interp, int objc,
+				 Tcl_Obj *CONST objv[])
+{
+	Pg_ConnectionId *connid;
+	PGconn	   *conn;
+	char	   *connString;
+	int         errorCode;
+
+	if (objc != 2)
+	{
+		Tcl_WrongNumArgs(interp, 1, objv, "connection");
+		return TCL_ERROR;
+	}
+
+	connString = Tcl_GetString(objv[1]);
+
+	conn = PgGetConnectionId(interp, connString, &connid);
+	if (conn == NULL)
+		return TCL_ERROR;
+
+	if (PgEndCopy(connid, &errorCode, 1) == -1) {
+		char *errorMessage = "I/O Error";
+		if(errorCode == EBUSY) {
+			errorMessage = "Busy";
+		}
+		Tcl_SetObjResult(interp, Tcl_NewStringObj(errorMessage, -1));
+		return TCL_ERROR;
+	}
+
+	return TCL_OK;
+}
+
+/**********************************
+ * PgCheckConnectionState(Pg_ConnectionId *connid)
+ *
+ * Called after an error from libpq that might have closed a connection
+ * (probably all of them). Unregister the notifier_channel and clear it out.
+ *
+ **********************************/
+int PgCheckConnectionState(Pg_ConnectionId *connid)
+{
+	// We don't have a connection, we can't do anything.
+	if(!connid->conn) {
+		return TCL_ERROR;
+	}
+
+	// Connection is still good, we're good.
+	if(PQstatus(connid->conn) != CONNECTION_BAD) {
+		return TCL_OK;
+	}
+
+	// Clean up notifiers and queue a connection loss event.
+	PgConnLossTransferEvents(connid);
+
+	return TCL_ERROR;
+}
