@@ -13,6 +13,9 @@
 
 #include <sqlite3.h>
 
+//#define IFDEBUG(thing) thing
+#define IFDEBUG(thing) /* */
+
 #define LAPPEND_STRING(i, o, s) Tcl_ListObjAppendElement((i), (o), Tcl_NewStringObj((s), -1));
 
 // From tclsqlite.c, part 1 of the hack, sqlite3 conveniently guarantees that the first element in
@@ -62,7 +65,7 @@ int
 Pg_sqlite_begin(Tcl_Interp *interp, sqlite3 *sqlite_db)
 {
 	char *errMsg;
-	if(sqlite3_exec(sqlite_db, "begin;", NULL, NULL, &errMsg) != SQLITE_OK) {
+	if(sqlite3_exec(sqlite_db, "savepoint pg_sqlite;", NULL, NULL, &errMsg) != SQLITE_OK) {
                 Tcl_AppendResult(interp, errMsg, " when beginning a transaction", (char *)NULL);
                 return TCL_ERROR;
         }
@@ -73,7 +76,7 @@ int
 Pg_sqlite_commit(Tcl_Interp *interp, sqlite3 *sqlite_db)
 {
 	char *errMsg;
-	if(sqlite3_exec(sqlite_db, "commit;", NULL, NULL, &errMsg) != SQLITE_OK) {
+	if(sqlite3_exec(sqlite_db, "release savepoint pg_sqlite;", NULL, NULL, &errMsg) != SQLITE_OK) {
                 Tcl_AppendResult(interp, errMsg, " when comitting a transaction", (char *)NULL);
                 return TCL_ERROR;
         }
@@ -356,7 +359,7 @@ Pg_sqlite_generateCheck(Tcl_Interp *interp, sqlite3 *sqlite_db, char *tableName,
 			strcpy(primaryKeyNames[k], column);
 		}
 		if(k != 0)
-			Tcl_AppendStringsToObj(where, ", ", (char *)NULL);
+			Tcl_AppendStringsToObj(where, " AND ", (char *)NULL);
 		Tcl_AppendStringsToObj(where, primaryKeyNames[k], " = ?", (char *)NULL);
 	}
 
@@ -401,6 +404,7 @@ Pg_sqlite_generateCheck(Tcl_Interp *interp, sqlite3 *sqlite_db, char *tableName,
 	// combine select list and where clause into statement, and turn it into a string
 	Tcl_AppendStringsToObj(sql, " FROM ", tableName, " WHERE (", Tcl_GetString(where), ");", (char *)NULL);
 
+	IFDEBUG(fprintf(stderr, "Generated check statement = '%s'\n", Tcl_GetString(sql)));
 	// create statement
 	if(Pg_sqlite_prepare(interp, sqlite_db, Tcl_GetString(sql), &statement) != TCL_OK) {
 		goto cleanup_and_exit;
@@ -462,6 +466,7 @@ Pg_sqlite_executeCheck(Tcl_Interp *interp, sqlite3 *sqlite_db, sqlite3_stmt *sta
 		const char       *errorMessage;
 
 		if (Pg_sqlite_bindValue(sqlite_db, statement, i, value, type, &errorMessage) != TCL_OK) {
+			IFDEBUG(fprintf(stderr, "ERROR binding primary key(%d) value = '%s' type = %d\n", i, value, type));
 			Tcl_AppendResult(interp, errorMessage, (char *)NULL);
 			goto cleanup_and_exit;
 		}
@@ -717,9 +722,9 @@ Pg_sqlite_probe(Tcl_Interp *interp, Tcl_ObjCmdProc **procPtr)
 	static Tcl_ObjCmdProc *sqlite3_ObjProc = NULL;
 
 	if (sqlite3_ObjProc == NULL) {
-		char               cmd_name[256];
-		char               create_cmd[256];
-		char               delete_cmd[256];
+		char               cmd_name[256 + 1];
+		char               create_cmd[256 + 18 + 1];
+		char               delete_cmd[256 + 7 + 1];
 		struct Tcl_CmdInfo cmd_info;
 
 		if (Tcl_Eval(interp, "package require sqlite3") != TCL_OK) {
@@ -727,8 +732,8 @@ Pg_sqlite_probe(Tcl_Interp *interp, Tcl_ObjCmdProc **procPtr)
 		}
 
 		snprintf(cmd_name, 256, "::dummy%d", getpid());
-		snprintf(create_cmd, 256, "sqlite3 %s :memory:", cmd_name);
-		snprintf(delete_cmd, 256, "%s close", cmd_name);
+		snprintf(create_cmd, 256 + 18, "sqlite3 %s :memory:", cmd_name);
+		snprintf(delete_cmd, 256 + 7, "%s close", cmd_name);
 
 		if (Tcl_Eval(interp, create_cmd) != TCL_OK) {
 			return TCL_ERROR;
@@ -1672,9 +1677,11 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 			int      maxValid = 0;
 			PGconn   *conn = NULL;
 			PGresult *result = NULL;
+			Pg_resultid *resultid;
+			Pg_ConnectionId *connid;
 
 			if(rowbyrow) {
-				conn = PgGetConnectionId(interp, pghandle_name, NULL);
+				conn = PgGetConnectionId(interp, pghandle_name, &connid);
 				if(conn == NULL) {
 					Tcl_AppendResult (interp, " while getting connection from ", pghandle_name, (char *)NULL);
 					returnCode = TCL_ERROR;
@@ -1683,7 +1690,8 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 				PQsetSingleRowMode(conn);
 				result = PQgetResult(conn);
 			} else {
-				result = PgGetResultId(interp, pghandle_name, NULL);
+				result = PgGetResultId(interp, pghandle_name, &resultid);
+				connid = resultid->connid;
 			}
 
 			if(!result) {
@@ -1711,6 +1719,7 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 					errorMessage = PQresultErrorMessage(result);
 					if (!*errorMessage)
 						errorMessage = PQresStatus(status);
+					PgCheckConnectionState(connid);
 					returnCode = TCL_ERROR;
 					break;
 				}
@@ -1789,6 +1798,8 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 					if (sqlite3_step(statement) != SQLITE_DONE) {
 						errorMessage = sqlite3_errmsg(sqlite_db);
 						returnCode = TCL_ERROR;
+						if(rowbyrow)
+							PgCheckConnectionState(connid);
 						goto import_cleanup_and_exit;
 					}
 					sqlite3_reset(statement);
@@ -1812,6 +1823,11 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 				if(rowbyrow) {
 					PQclear(result);
 					result = PQgetResult(conn);
+					if(!result && PgCheckConnectionState(connid) != TCL_OK) {
+						errorMessage = "CONNECTION_BAD";
+						returnCode = TCL_ERROR;;
+						goto import_cleanup_and_exit;
+					}
 				} else {
 					result = NULL;
 				}
@@ -1825,6 +1841,10 @@ Pg_sqlite(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 					if(!conn) break;
 					result = PQgetResult(conn);
 				}
+				if(PgCheckConnectionState(connid) != TCL_OK) {
+					errorMessage = "CONNECTION_BAD";
+					returnCode = TCL_ERROR;;
+				}
 			}
 
 			if(statement) {
